#  Copyright 2023 Hathor Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from copy import deepcopy
from typing import Any, TypeVar

K = TypeVar('K')


def deep_merge(first_dict: dict[K, Any], second_dict: dict[K, Any]) -> dict[K, Any]:
    """
    Recursively merges two dicts, returning a new one with the merged values. Keeps both input dicts intact.

    Note: will raise RecursionError if there's a circular reference in both dicts.

    >>> dict1 = dict(a=1, b=dict(c=2, d=3), e=dict(f=4))
    >>> dict2 = dict(b=dict(d=5, e=6), e=7)
    >>> result = deep_merge(dict1, dict2)
    >>> result == dict(a=1, b=dict(c=2, d=5, e=6), e=7)
    True
    >>> dict1 == dict(a=1, b=dict(c=2, d=3), e=dict(f=4))
    True
    >>> dict2 == dict(b=dict(d=5, e=6), e=7)
    True
    """
    merged = deepcopy(first_dict)

    def do_deep_merge(first: dict[K, Any], second: dict[K, Any]) -> dict[K, Any]:
        for key in second:
            if key in first and isinstance(first[key], dict) and isinstance(second[key], dict):
                do_deep_merge(first[key], second[key])
            else:
                first[key] = second[key]

        return first

    return do_deep_merge(merged, second_dict)



import os
from pathlib import Path
from typing import NamedTuple, Optional

from hathorlib.conf.settings import HathorSettings as Settings
from hathorlib.conf.utils import load_module_settings, load_yaml_settings

_config_file = None


class _SettingsMetadata(NamedTuple):
    source: str
    is_yaml: bool
    settings: Settings


_settings_singleton: Optional[_SettingsMetadata] = None


def HathorSettings() -> Settings:
    """ Return configuration file namedtuple
        Get the file from environment variable 'TXMINING_CONFIG_FILE'
        If not set we return the config file of the mainnet
    """
    settings_module_filepath = os.environ.get('HATHOR_CONFIG_FILE')
    if settings_module_filepath is not None:
        return _load_settings_singleton(settings_module_filepath, is_yaml=False)

    default_settings = str(Path(__file__).parent / 'mainnet.yml')
    settings_yaml_filepath = os.environ.get('HATHOR_CONFIG_YAML', default_settings)
    return _load_settings_singleton(settings_yaml_filepath, is_yaml=True)


def _load_settings_singleton(source: str, *, is_yaml: bool) -> Settings:
    global _settings_singleton

    if _settings_singleton is not None:
        if _settings_singleton.is_yaml != is_yaml:
            raise Exception('loading config twice with a different file type')
        if _settings_singleton.source != source:
            raise Exception('loading config twice with a different file')

        return _settings_singleton.settings

    settings_loader = load_yaml_settings if is_yaml else load_module_settings
    _settings_singleton = _SettingsMetadata(
        source=source,
        is_yaml=is_yaml,
        settings=settings_loader(Settings, source)
    )

    return _settings_singleton.settings



"""
Copyright (c) Hathor Labs and its affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

import os
from enum import StrEnum, auto, unique
from math import log
from typing import Annotated, Optional

from pydantic import BaseModel, BeforeValidator, ConfigDict, computed_field, field_validator, model_validator
from typing_extensions import Self

from hathorlib.conf.utils import parse_hex_str


@unique
class FeatureSetting(StrEnum):
    """Enum to configure the state of a feature."""

    # Completely disabled.
    DISABLED = auto()

    # Completely enabled since network creation.
    ENABLED = auto()

    # Enabled through Feature Activation.
    FEATURE_ACTIVATION = auto()

    def __bool__(self) -> bool:
        """
        >>> bool(FeatureSetting.DISABLED)
        False
        >>> bool(FeatureSetting.ENABLED)
        True
        >>> bool(FeatureSetting.FEATURE_ACTIVATION)
        True
        """
        return self in (FeatureSetting.ENABLED, FeatureSetting.FEATURE_ACTIVATION)


class HathorSettings(BaseModel):
    model_config = ConfigDict(extra='ignore')

    # Name of the network: "mainnet", "testnet-alpha", "testnet-bravo", ...
    NETWORK_NAME: str

    # Version byte of the address in P2PKH
    P2PKH_VERSION_BYTE: Annotated[bytes, BeforeValidator(parse_hex_str)]

    # Version byte of the address in MultiSig
    MULTISIG_VERSION_BYTE: Annotated[bytes, BeforeValidator(parse_hex_str)]

    # HTR Token UID
    HATHOR_TOKEN_UID: bytes = b'\x00'

    # Maximum number of characters in a token name
    MAX_LENGTH_TOKEN_NAME: int = 30

    # Maximum number of characters in a token symbol
    MAX_LENGTH_TOKEN_SYMBOL: int = 5

    # Name of the Hathor token
    HATHOR_TOKEN_NAME: str = 'Hathor'

    # Symbol of the Hathor token
    HATHOR_TOKEN_SYMBOL: str = 'HTR'

    # The name and symbol of the native token. This is only used in APIs to serve clients.
    NATIVE_TOKEN_NAME: str = 'Hathor'
    NATIVE_TOKEN_SYMBOL: str = 'HTR'

    # Number of decimal places for the Hathor token
    DECIMAL_PLACES: int = 2

    # Minimum weight of a tx
    MIN_TX_WEIGHT: int = 14

    # Multiplier coefficient to adjust the minimum weight of a normal tx to 18
    MIN_TX_WEIGHT_COEFFICIENT: float = 1.6

    # Amount in which tx min weight reaches the middle point between the minimum and maximum weight
    MIN_TX_WEIGHT_K: int = 100

    # Maximum size of the tx output's script allowed for a tx to be standard
    PUSHTX_MAX_OUTPUT_SCRIPT_SIZE: int = 256

    # Maximum number of tx outputs of Data Script type
    MAX_DATA_SCRIPT_OUTPUTS: int = 25

    # Max length in bytes allowed for on-chain blueprint code after decompression, 240KB (not KiB)
    NC_ON_CHAIN_BLUEPRINT_CODE_MAX_SIZE_UNCOMPRESSED: int = 240_000

    # Max length in bytes allowed for on-chain blueprint code inside the transaction, 24KB (not KiB)
    NC_ON_CHAIN_BLUEPRINT_CODE_MAX_SIZE_COMPRESSED: int = 24_000

    # Initial bootstrap servers
    BOOTSTRAP_DNS: list[str] = []

    # enable peer whitelist
    ENABLE_PEER_WHITELIST: bool = False

    # weather to use the whitelist with sync-v2 peers, does not affect whether the whitelist is enabled or not, it will
    # always be enabled for sync-v1 if it is enabled
    USE_PEER_WHITELIST_ON_SYNC_V2: bool = True

    # Genesis pre-mined tokens
    GENESIS_TOKEN_UNITS: int = 1 * (10 ** 9)  # 1B

    GENESIS_TOKENS: int = 1 * (10 ** 9) * (10 ** 2)  # 100B = GENESIS_TOKEN_UNITS * (10 ** DECIMAL_PLACES)

    # Fee rate settings
    FEE_PER_OUTPUT: int = 1

    @computed_field  # type: ignore[prop-decorator]
    @property
    def FEE_DIVISOR(self) -> int:
        """Divisor used for evaluating fee amounts"""
        result = 1 / self.TOKEN_DEPOSIT_PERCENTAGE
        assert result.is_integer()
        return int(result)

    # To disable reward halving, just set this to `None` and make sure that INITIAL_TOKEN_UNITS_PER_BLOCK is equal to
    # MINIMUM_TOKEN_UNITS_PER_BLOCK.
    BLOCKS_PER_HALVING: Optional[int] = 2 * 60 * 24 * 365  # 1051200, every 365 days

    INITIAL_TOKEN_UNITS_PER_BLOCK: int = 64
    MINIMUM_TOKEN_UNITS_PER_BLOCK: int = 8

    @computed_field  # type: ignore[prop-decorator]
    @property
    def INITIAL_TOKENS_PER_BLOCK(self) -> int:
        return int(self.INITIAL_TOKEN_UNITS_PER_BLOCK * (10 ** self.DECIMAL_PLACES))

    @computed_field  # type: ignore[prop-decorator]
    @property
    def MINIMUM_TOKENS_PER_BLOCK(self) -> int:
        return int(self.MINIMUM_TOKEN_UNITS_PER_BLOCK * (10 ** self.DECIMAL_PLACES))

    # Assume that: amount < minimum
    # But, amount = initial / (2**n), where n = number_of_halvings. Thus:
    #   initial / (2**n) < minimum
    #   initial / minimum < 2**n
    #   2**n > initial / minimum
    # Applying log to both sides:
    #   n > log2(initial / minimum)
    #   n > log2(initial) - log2(minimum)
    @computed_field  # type: ignore[prop-decorator]
    @property
    def MAXIMUM_NUMBER_OF_HALVINGS(self) -> int:
        return int(log(self.INITIAL_TOKEN_UNITS_PER_BLOCK, 2) - log(self.MINIMUM_TOKEN_UNITS_PER_BLOCK, 2))

    # Average time between blocks.
    AVG_TIME_BETWEEN_BLOCKS: int = 30  # in seconds

    # Genesis pre-mined outputs
    # P2PKH HMcJymyctyhnWsWTXqhP9txDwgNZaMWf42
    #
    # To generate a new P2PKH script, run:
    # >>> from hathor.transaction.scripts import P2PKH
    # >>> import base58
    # >>> address = base58.b58decode('HMcJymyctyhnWsWTXqhP9txDwgNZaMWf42')
    # >>> P2PKH.create_output_script(address=address).hex()
    GENESIS_OUTPUT_SCRIPT: Annotated[bytes, BeforeValidator(parse_hex_str)] = (
        bytes.fromhex("76a914a584cf48b161e4a49223ed220df30037ab740e0088ac")
    )

    # Genesis timestamps, nonces and hashes

    # Timestamp used for the genesis block
    GENESIS_BLOCK_TIMESTAMP: int = 1572636343

    @computed_field  # type: ignore[prop-decorator]
    @property
    def GENESIS_TX1_TIMESTAMP(self) -> int:
        """Timestamp used for the first genesis transaction."""
        return self.GENESIS_BLOCK_TIMESTAMP + 1

    @computed_field  # type: ignore[prop-decorator]
    @property
    def GENESIS_TX2_TIMESTAMP(self) -> int:
        """Timestamp used for the second genesis transaction."""
        return self.GENESIS_BLOCK_TIMESTAMP + 2

    GENESIS_BLOCK_NONCE: int = 3526202
    GENESIS_BLOCK_HASH: Annotated[bytes, BeforeValidator(parse_hex_str)] = (
        bytes.fromhex(
            "000007eb968a6cdf0499e2d033faf1e163e0dc9cf41876acad4d421836972038"
        )
    )
    GENESIS_TX1_NONCE: int = 12595
    GENESIS_TX1_HASH: Annotated[bytes, BeforeValidator(parse_hex_str)] = bytes.fromhex(
        "00025d75e44804a6a6a099f4320471c864b38d37b79b496ee26080a2a1fd5b7b"
    )
    GENESIS_TX2_NONCE: int = 21301
    GENESIS_TX2_HASH: Annotated[bytes, BeforeValidator(parse_hex_str)] = bytes.fromhex(
        "0002c187ab30d4f61c11a5dc43240bdf92dba4d19f40f1e883b0a5fdac54ef53"
    )

    # Weight of genesis and minimum weight of a tx/block
    MIN_BLOCK_WEIGHT: int = 21
    MIN_SHARE_WEIGHT: int = 21

    # Maximum distance between two consecutive blocks (in seconds), except for genesis.
    # This prevent some DoS attacks exploiting the calculation of the score of a side chain.
    #     P(t > T) = exp(-MAX_DISTANCE_BETWEEN_BLOCKS / AVG_TIME_BETWEEN_BLOCKS)
    #     P(t > T) = exp(-35) = 6.3051e-16
    MAX_DISTANCE_BETWEEN_BLOCKS: int = 150 * 30  # AVG_TIME_BETWEEN_BLOCKS

    # Enable/disable weight decay.
    WEIGHT_DECAY_ENABLED: bool = True

    # Minimum distance between two consecutive blocks that enables weight decay.
    # Assuming that the hashrate is constant, the probability of activating is:
    #     P(t > T) = exp(-WEIGHT_DECAY_ACTIVATE_DISTANCE / AVG_TIME_BETWEEN_BLOCKS)
    #     P(t > T) = exp(-120) = 7.66e-53
    # But, if the hashrate drops 40 times, the expected time to find the next block
    # becomes 40 * AVG_TIME_BETWEEN_BLOCKS = 20 minutes and the probability of
    # activating the decay is exp(-3) = 0.05 = 5%.
    WEIGHT_DECAY_ACTIVATE_DISTANCE: int = 120 * 30  # AVG_TIME_BETWEEN_BLOCKS

    # Window size of steps in which the weight is reduced when decaying is activated.
    # The maximum number of steps is:
    #     max_steps = floor((MAX_DISTANCE_BETWEEN_BLOCKS - WEIGHT_DECAY_ACTIVATE_DISTANCE) / WEIGHT_DECAY_WINDOW_SIZE)
    # Using these parameters, `max_steps = 15`.
    WEIGHT_DECAY_WINDOW_SIZE: int = 60

    # Amount to reduce the weight when decaying is activated.
    #     adj_weight = weight - decay
    #     difficulty = 2**adj_weight
    #     difficulty = 2**(weight - decay)
    #     difficulty = 2**weight / 2**decay
    # As 2**(-2.73) = 0.15072, it reduces the mining difficulty for 15% of the original weight.
    # Finally, the maximum decay is `max_steps * WEIGHT_DECAY_AMOUNT`.
    # As `max_steps = 15`, then `max_decay = 2**(-15 * 2.73) = 4.71e-13`.
    WEIGHT_DECAY_AMOUNT: float = 2.73

    # Number of blocks to be found with the same hash algorithm as `block`.
    # The bigger it is, the smaller the variance of the hash rate estimator is.
    BLOCK_DIFFICULTY_N_BLOCKS: int = 134

    # Size limit in bytes for Block data field
    BLOCK_DATA_MAX_SIZE: int = 100

    # Number of subfolders in the storage folder (used in JSONStorage and CompactStorage)
    STORAGE_SUBFOLDERS: int = 256

    # Maximum level of the neighborhood graph generated by graphviz
    MAX_GRAPH_LEVEL: int = 3

    # Maximum difference between our latest timestamp and a peer's synced timestamp to consider
    # that the peer is synced (in seconds).
    P2P_SYNC_THRESHOLD: int = 60

    # This multiplier will be used to decide whether the fullnode has had recent activity in the p2p sync.
    # This info will be used in the readiness endpoint as one of the checks.
    #
    # We will multiply it by the AVG_TIME_BETWEEN_BLOCKS, and compare the result with the gap between the
    # current time and the latest timestamp in the database.
    #
    # If the gap is bigger than the calculated threshold, than we will say the fullnode is not ready (unhealthy).
    #
    # Using (P2P_RECENT_ACTIVITY_THRESHOLD_MULTIPLIER * AVG_TIME_BETWEEN_BLOCKS) as threshold will have false
    # positives.
    # The probability of a false positive is exp(-N), assuming the hash rate is constant during the period.
    # In other words, a false positive is likely to occur every exp(N) blocks. If the hash rate decreases
    # quickly, this probability gets bigger.
    #
    # For instance, P2P_RECENT_ACTIVITY_THRESHOLD_MULTIPLIER=5 would get a false positive every 90 minutes.
    # For P2P_RECENT_ACTIVITY_THRESHOLD_MULTIPLIER=10, we would have a false positive every 8 days.
    # For P2P_RECENT_ACTIVITY_THRESHOLD_MULTIPLIER=15, we would have a false positive every 3 years.
    #
    # On the other side, using higher numbers will lead to a higher delay for the fullnode to start reporting
    # as not ready.
    #
    # So for use cases that may need more reponsiveness on this readiness check, at the cost of some eventual false
    # positive, it could be a good idea to decrease the value in this setting.
    P2P_RECENT_ACTIVITY_THRESHOLD_MULTIPLIER: int = 15

    # Whether to warn the other peer of the reason for closing the connection
    WHITELIST_WARN_BLOCKED_PEERS: bool = False

    # Maximum number of opened threads that are solving POW for send tokens
    MAX_POW_THREADS: int = 5

    # The error tolerance, to allow small rounding errors in Python, when comparing weights,
    # accumulated weights, and scores
    # How to use:
    # if abs(w1 - w2) < WEIGHT_TOL:
    #     print('w1 and w2 are equal')

    # if w1 < w2 - WEIGHT_TOL:
    #     print('w1 is smaller than w2')

    # if w1 <= w2 + WEIGHT_TOL:
    #     print('w1 is smaller than or equal to w2')

    # if w1 > w2 + WEIGHT_TOL:
    #     print('w1 is greater than w2')

    # if w1 >= w2 - WEIGHT_TOL:
    #     print('w1 is greater than or equal to w2')
    WEIGHT_TOL: float = 1e-10

    # Maximum difference between the weight and the min_weight.
    MAX_TX_WEIGHT_DIFF: float = 4.0
    MAX_TX_WEIGHT_DIFF_ACTIVATION: float = 32.0

    # Maximum number of txs or blocks (each, not combined) to show on the dashboard
    MAX_DASHBOARD_COUNT: int = 15

    # Maximum number of txs or blocks returned by the '/transaction' endpoint
    MAX_TX_COUNT: int = 15

    # URL prefix where API is served, for instance: /v1a/status
    API_VERSION_PREFIX: str = 'v1a'

    # If should use stratum to resolve pow of transactions in send tokens resource
    SEND_TOKENS_STRATUM: bool = True

    # Maximum number of subscribed addresses per websocket connection
    WS_MAX_SUBS_ADDRS_CONN: Optional[int] = None

    # Maximum number of subscribed addresses that do not have any outputs (also per websocket connection)
    WS_MAX_SUBS_ADDRS_EMPTY: Optional[int] = None

    # Whether miners are assumed to mine txs by default
    STRATUM_MINE_TXS_DEFAULT: bool = True

    # Percentage used to calculate the number of HTR that must be deposited when minting new tokens
    # The same percentage is used to calculate the number of HTR that must be withdraw when melting tokens
    # See for further information, see [rfc 0011-token-deposit].
    TOKEN_DEPOSIT_PERCENTAGE: float = 0.01

    @field_validator('TOKEN_DEPOSIT_PERCENTAGE', mode='after')
    @classmethod
    def _validate_token_deposit_percentage(cls, token_deposit_percentage: float) -> float:
        """Validate that TOKEN_DEPOSIT_PERCENTAGE results in an integer FEE_DIVISOR."""
        result = 1 / token_deposit_percentage
        if not result.is_integer():
            raise ValueError(
                f'TOKEN_DEPOSIT_PERCENTAGE must result in an integer FEE_DIVISOR. '
                f'Got TOKEN_DEPOSIT_PERCENTAGE={token_deposit_percentage}, FEE_DIVISOR={result}'
            )
        return token_deposit_percentage

    # Array with the settings parameters that are used when calculating the settings hash
    P2P_SETTINGS_HASH_FIELDS: list[str] = [
        'P2PKH_VERSION_BYTE',
        'MULTISIG_VERSION_BYTE',
        'MIN_BLOCK_WEIGHT',
        'MIN_TX_WEIGHT',
        'BLOCK_DATA_MAX_SIZE'
    ]

    # Maximum difference allowed between current time and a received tx timestamp (in seconds). Also used
    # during peer connection. Peers shouldn't have their clocks more than MAX_FUTURE_TIMESTAMP_ALLOWED/2 apart
    MAX_FUTURE_TIMESTAMP_ALLOWED: int = 5 * 60

    # Multiplier for the value to increase the timestamp for the next retry moment to connect to the peer
    PEER_CONNECTION_RETRY_INTERVAL_MULTIPLIER: int = 5

    # Maximum retry interval for retrying to connect to the peer
    PEER_CONNECTION_RETRY_MAX_RETRY_INTERVAL: int = 300

    # Number max of connections in the p2p network
    PEER_MAX_CONNECTIONS: int = 125

    # Maximum period without receiving any messages from ther peer (in seconds).
    PEER_IDLE_TIMEOUT: int = 60

    # Maximum number of entrypoints that we accept that a peer broadcasts
    PEER_MAX_ENTRYPOINTS: int = 30

    # Filepath of ca certificate file to generate connection certificates
    CA_FILEPATH: str = os.path.join(os.path.dirname(__file__), '../cert/ca.crt')

    # Filepath of ca key file to sign connection certificates
    CA_KEY_FILEPATH: str = os.path.join(os.path.dirname(__file__), '../cert/ca.key')

    # Timeout (in seconds) for the downloading deferred (in the downloader) when syncing two peers
    GET_DATA_TIMEOUT: int = 90

    # Number of retries for downloading a tx from a peer (in the downloader)
    GET_DATA_RETRIES: int = 5

    # After how many blocks can a reward be spent
    REWARD_SPEND_MIN_BLOCKS: int = 300

    # Mamimum number of inputs accepted
    MAX_NUM_INPUTS: int = 255

    # Mamimum number of outputs accepted
    MAX_NUM_OUTPUTS: int = 255

    # Maximum size of each txout's script (in bytes)
    MAX_OUTPUT_SCRIPT_SIZE: int = 1024

    # Maximum size of each txin's data (in bytes)
    MAX_INPUT_DATA_SIZE: int = 1024

    # Maximum number of pubkeys per OP_CHECKMULTISIG
    MAX_MULTISIG_PUBKEYS: int = 20

    # Maximum number of signatures per OP_CHECKMULTISIG
    MAX_MULTISIG_SIGNATURES: int = 15

    # Maximum number of sig operations of all inputs on a given tx
    # including the redeemScript in case of MultiSig
    MAX_TX_SIGOPS_INPUT: int = 255 * 5

    # Maximum number of sig operations of all outputs on a given tx
    MAX_TX_SIGOPS_OUTPUT: int = 255 * 5

    # Maximum number of transactions returned on addresses history API
    MAX_TX_ADDRESSES_HISTORY: int = 150

    # Maximum number of elements (inputs and outputs) to be returned on address history API
    # As a normal tx has ~2-4 inputs and 2 outputs, I would say the maximum should be 150*6 = 900 elements
    MAX_INPUTS_OUTPUTS_ADDRESS_HISTORY: int = 6 * 150  # MAX_TX_ADDRESSES_HISTORY

    # Maximum number of TXs that will be sent by the Mempool API.
    MEMPOOL_API_TX_LIMIT: int = 100

    # Capabilities
    CAPABILITY_WHITELIST: str = 'whitelist'
    CAPABILITY_SYNC_VERSION: str = 'sync-version'
    CAPABILITY_GET_BEST_BLOCKCHAIN: str = 'get-best-blockchain'
    CAPABILITY_IPV6: str = 'ipv6'  # peers announcing this capability will be relayed ipv6 entrypoints from other peers
    CAPABILITY_NANO_STATE: str = 'nano-state'  # indicates support for nano-state commands

    # Where to download whitelist from
    WHITELIST_URL: Optional[str] = None

    # Interval (in seconds) to broadcast dashboard metrics to websocket connections
    WS_SEND_METRICS_INTERVAL: int = 1

    # Interval (in seconds) to write data to prometheus
    PROMETHEUS_WRITE_INTERVAL: int = 15

    # Interval (in seconds) to update GC data for prometheus
    PROMETHEUS_UPDATE_GC_INTERVAL: int = 60

    # Interval (in seconds) to collect metrics data
    METRICS_COLLECT_DATA_INTERVAL: int = 5

    # Interval (in seconds) to collect metrics data from rocksdb
    METRICS_COLLECT_ROCKSDB_DATA_INTERVAL: int = 86400  # 1 day

    # Used on testing to enable slow asserts that help catch bugs but we don't want to run in production
    SLOW_ASSERTS: bool = False

    # List of soft voided transaction.
    SOFT_VOIDED_TX_IDS: list[Annotated[bytes, BeforeValidator(parse_hex_str)]] = []

    # List of transactions to skip verification.
    SKIP_VERIFICATION: list[Annotated[bytes, BeforeValidator(parse_hex_str)]] = []

    # Identifier used in metadata's voided_by to mark a tx as soft-voided.
    SOFT_VOIDED_ID: bytes = b'tx-non-grata'

    # Identifier used in metadata's voided_by when an unexpected exception occurs at consensus.
    CONSENSUS_FAIL_ID: bytes = b'consensus-fail'

    # Identifier used in metadata's voided_by to mark a tx as partially validated.
    PARTIALLY_VALIDATED_ID: bytes = b'pending-validation'

    # Maximum number of sync running simultaneously.
    MAX_ENABLED_SYNC: int = 8

    # Time to update the peers that are running sync.
    SYNC_UPDATE_INTERVAL: int = 10 * 60  # seconds

    # Interval to re-run peer discovery.
    PEER_DISCOVERY_INTERVAL: int = 5 * 60  # seconds

    # Maximum number of GET_TIPS delayed calls per connection while running sync.
    MAX_GET_TIPS_DELAYED_CALLS: int = 5

    # Maximum number of blocks in the best blockchain list.
    MAX_BEST_BLOCKCHAIN_BLOCKS: int = 20

    # Default number of blocks in the best blockchain list.
    DEFAULT_BEST_BLOCKCHAIN_BLOCKS: int = 10

    # Time in seconds to request the best blockchain from peers.
    BEST_BLOCKCHAIN_INTERVAL: int = 5  # seconds

    # Merged mining settings. The old value is going to be replaced by the new value through Feature Activation.
    OLD_MAX_MERKLE_PATH_LENGTH: int = 12
    NEW_MAX_MERKLE_PATH_LENGTH: int = 20

    # Maximum number of tx tips to accept in the initial phase of the mempool sync 1000 is arbitrary, but it should be
    # more than enough for the forseeable future
    MAX_MEMPOOL_RECEIVING_TIPS: int = 1000

    # Max number of peers simultanously stored in the node
    MAX_VERIFIED_PEERS: int = 10_000

    # Max number of peers simultanously stored per-connection
    MAX_UNVERIFIED_PEERS_PER_CONN: int = 100

    # Used to enable nano contracts.
    ENABLE_NANO_CONTRACTS: FeatureSetting = FeatureSetting.DISABLED

    # Used to enable fee-based tokens.
    ENABLE_FEE_BASED_TOKENS: FeatureSetting = FeatureSetting.DISABLED

    # Used to enable opcodes V2.
    ENABLE_OPCODES_V2: FeatureSetting = FeatureSetting.DISABLED

    # List of enabled blueprints.
    BLUEPRINTS: dict[bytes, str] = {}

    @field_validator('BLUEPRINTS', mode='before')
    @classmethod
    def _parse_blueprints(cls, blueprints_raw: dict[str, str]) -> dict[bytes, str]:
        """Parse dict[str, str] into dict[bytes, str]."""
        blueprints: dict[bytes, str] = {}
        for _id_str, _name in blueprints_raw.items():
            _id = bytes.fromhex(_id_str)
            if _id in blueprints:
                raise TypeError(f'Duplicate blueprint id: {_id_str}')
            blueprints[_id] = _name
        return blueprints

    # The pubkeys allowed to create on-chain-blueprints in the network
    # XXX: in the future this restriction will be lifted, possibly through a feature activation
    NC_ON_CHAIN_BLUEPRINT_RESTRICTED: bool = True
    NC_ON_CHAIN_BLUEPRINT_ALLOWED_ADDRESSES: list[str] = []

    # TODO: align this with a realistic value later
    # fuel units are arbitrary but it's roughly the number of opcodes, memory_limit is in bytes
    NC_INITIAL_FUEL_TO_LOAD_BLUEPRINT_MODULE: int = 100_000  # 100K opcodes
    NC_MEMORY_LIMIT_TO_LOAD_BLUEPRINT_MODULE: int = 100 * 1024 * 1024  # 100MiB
    NC_INITIAL_FUEL_TO_CALL_METHOD: int = 1_000_000  # 1M opcodes
    NC_MEMORY_LIMIT_TO_CALL_METHOD: int = 1024 * 1024 * 1024  # 1GiB

    @model_validator(mode='after')
    def _validate_genesis_tokens(self) -> Self:
        """Validate genesis tokens."""
        genesis_tokens = self.GENESIS_TOKENS
        genesis_token_units = self.GENESIS_TOKEN_UNITS
        decimal_places = self.DECIMAL_PLACES

        if genesis_tokens != genesis_token_units * (10 ** decimal_places):
            raise ValueError(
                f'invalid tokens: GENESIS_TOKENS={genesis_tokens}, '
                f'GENESIS_TOKEN_UNITS={genesis_token_units}, DECIMAL_PLACES={decimal_places}'
            )
        return self



"""
Copyright (c) Hathor Labs and its affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""
import importlib
from pathlib import Path
from typing import TypeVar, Union

from pydantic import BaseModel

T = TypeVar('T', bound=BaseModel)


def parse_hex_str(hex_str: Union[str, bytes]) -> bytes:
    """Parse a raw hex string into bytes."""
    if isinstance(hex_str, str):
        return bytes.fromhex(hex_str.lstrip('x'))

    if not isinstance(hex_str, bytes):
        raise ValueError(f'expected \'str\' or \'bytes\', got {hex_str}')

    return hex_str


def load_yaml_settings(model: type[T], filepath: str) -> T:
    """
    Load a settings model (pydantic based) and a filepath to a yaml file and returns a validated instance.
    YAML settings may use the `extends` key to merge definition with another existing file.
    """
    from hathorlib.utils.yaml import model_from_extended_yaml
    return model_from_extended_yaml(model, filepath=filepath, custom_root=Path(__file__).parent)


def load_module_settings(model: type[T], module_path: str) -> T:
    """Load module"""
    settings_module = importlib.import_module(module_path)
    settings = getattr(settings_module, 'SETTINGS')
    return model.model_validate(settings)



# Copyright 2025 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from types import UnionType
from typing import Generic, TypeVar, get_args as _typing_get_args, get_origin as _typing_get_origin
from weakref import WeakValueDictionary

from typing_extensions import Self

T = TypeVar('T')


def get_origin(t: type | UnionType, /) -> type | None:
    """Extension of typing.get_origin to also work with classes that use InnerTypeMixin"""
    if isinstance(t, type) and issubclass(t, InnerTypeMixin):
        return getattr(t, '__origin__', None)
    return _typing_get_origin(t)


def get_args(t: type | UnionType, /) -> tuple[type, ...] | None:
    """Extension of typing.get_args to also work with classes that use InnerTypeMixin"""
    if isinstance(t, type) and issubclass(t, InnerTypeMixin):
        return getattr(t, '__args__', None)
    return _typing_get_args(t)


class InnerTypeMixin(Generic[T]):
    """
    Mixin class that exposes its single type‐argument at runtime as `self.__inner_type__`,
    enforces exactly one type argument at subscription time, caches parameterized subclasses
    so C[int] is C[int], and provides a clean repr listing public fields.

    >>> from typing import TypeVar
    >>> U = TypeVar('U')
    >>> class MyData(InnerTypeMixin, Generic[T]):
    ...     def __init__(self, data: T):
    ...         self.data = data
    ...
    >>> class MyCounter(InnerTypeMixin, Generic[T]):
    ...     def __init__(self, first: T, count: int):
    ...         self.first = first
    ...         self.count = count
    ...

    # 1) You must supply exactly one type argument:
    >>> try:
    ...     MyData(1)
    ... except TypeError as e:
    ...     print(e)
    MyData[...] requires exactly one type argument, got none

    >>> try:
    ...     MyData[int, str](1)
    ... except TypeError as e:
    ...     print(e)
    MyData[...] expects exactly one type argument; got 2

    # You may write MyData[U] for signatures, but instantiation will reject a bare TypeVar:
    >>> MyData[U]    # no error
    <class 'hathorlib.utils.typing.MyData'>

    >>> try:
    ...     MyData[U]()
    ... except TypeError as e:
    ...     print(e)
    MyData[...] requires a concrete type argument, got ~U

    # Correct usage with a concrete type:
    >>> sd = MyData[int](123)
    >>> MyData[int] is MyData[int]
    True
    >>> sd.__inner_type__ is int
    True
    >>> print(sd)
    MyData[int](data=123)

    # Works with multiple fields too:
    >>> h = MyCounter[str]("foo", 42)
    >>> h.__inner_type__ is str
    True
    >>> print(h)
    MyCounter[str](first='foo', count=42)
    """

    # cache shared by all subclasses, maps concrete inner_type -> subclass, but doesn't keep subclasses alive if it has
    # no live references anymore, this keeps the cache from growing indefinitely in case of dynamically generated
    # classes, there's no point in holding unreferenced classes here
    __type_cache: WeakValueDictionary[tuple[type, type], type[Self]] = WeakValueDictionary()

    # this class will expose this instance property
    __inner_type__: type[T]

    @classmethod
    def __extract_inner_type__(cls, args: tuple[type, ...], /) -> type[T]:
        """Defines how to convert the received argument tuples into the stored type.

        If customization is needed, this class method is the place to do it. I could be used so only the origin-type is
        stored, or to accept multiple arguments and store a tuple of types, or to convert the arguments into different
        types.
        """
        if len(args) != 1:
            raise TypeError(f'{cls.__name__}[...] expects exactly one type argument; got {len(args)}')
        inner_type, = args
        return inner_type

    @classmethod
    def __class_getitem__(cls, params):
        # parameterizing the mixin itself delegates to Generic
        if cls is InnerTypeMixin:
            return super().__class_getitem__(params)

        # normalize to a 1-tuple
        args = params if isinstance(params, tuple) else (params,)
        inner_type = cls.__extract_inner_type__(args)

        cache = cls.__type_cache
        key = (cls, inner_type)
        sub = cache.get(key)
        if sub is None:
            # subclass keeps the same name for clean repr
            sub = type(cls.__name__, (cls,), {})
            sub.__inner_type__ = inner_type
            sub.__origin__ = cls
            sub.__args__ = (inner_type,)
            sub.__module__ = cls.__module__
            sub.__type_cache = cache
            cache[key] = sub
        return sub

    def __new__(cls, *args, **kwargs):
        # reject unsubscripted class
        if not get_args(cls):
            raise TypeError(f'{cls.__name__}[...] requires exactly one type argument, got none')

        # reject if the subscribed‐in type is still a TypeVar
        inner_type = getattr(cls, '__inner_type__', None)
        if isinstance(inner_type, TypeVar):
            raise TypeError(f'{cls.__name__}[...] requires a concrete type argument, got {inner_type!r}')

        # build instance and copy down the inner type
        self = super().__new__(cls)
        self.__inner_type__ = inner_type
        return self

    def __repr__(self) -> str:
        name = type(self).__name__
        t = self.__inner_type__
        tname = getattr(t, '__name__', repr(t))
        public = [(n, v) for n, v in vars(self).items() if not n.startswith('_')]
        if public:
            body = ', '.join(f'{n}={v!r}' for n, v in public)
            return f'{name}[{tname}]({body})'
        return f'{name}[{tname}]()'


def is_subclass(cls: type, class_or_tuple: type | tuple[type] | UnionType, /) -> bool:
    """ Reimplements issubclass() with support for recursive NewType classes.

    Normal behavior from `issubclass`:

    >>> is_subclass(int, int)
    True
    >>> is_subclass(bool, int)
    True
    >>> is_subclass(bool, (int, str))
    True
    >>> is_subclass(bool, int | str)
    True
    >>> is_subclass(bool, bytes | str)
    False
    >>> is_subclass(str, int)
    False

    But `is_subclass` also works when a NewType is given as arg 1:

    >>> from typing import NewType
    >>> N = NewType('N', int)
    >>> is_subclass(N, int)
    True
    >>> is_subclass(N, int | str)
    True
    >>> is_subclass(N, str)
    False
    >>> M = NewType('M', N)
    >>> is_subclass(M, int)
    True
    >>> is_subclass(M, str)
    False
    >>> try:
    ...     is_subclass(M, N)
    ... except TypeError as e:
    ...     print(*e.args)
    issubclass() arg 2 must be a class, a tuple of classes, or a union

    It is also expeced to fail in the same way as `issubclass` when the resolving the NewType doesn't lead to a class:

    >>> F = NewType('F', 'not a class')
    >>> try:
    ...     is_subclass(F, str)
    ... except TypeError as e:
    ...     print(*e.args)
    issubclass() arg 1 must be a class
    """
    while (super_type := getattr(cls, '__supertype__', None)) is not None:
        cls = super_type
    return issubclass(cls, class_or_tuple)



# Copyright 2023 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hathorlib.nanocontracts.faux_immutable import (
    ALLOW_DUNDER_ATTR,
    ALLOW_INHERITANCE_ATTR,
    SKIP_VALIDATION_ATTR,
    FauxImmutable,
    FauxImmutableMeta,
    __set_faux_immutable__,
    create_with_shell,
)
from hathorlib.nanocontracts.nanocontract import DeprecatedNanoContract
from hathorlib.nanocontracts.on_chain_blueprint import OnChainBlueprint

__all__ = [
    'ALLOW_DUNDER_ATTR',
    'ALLOW_INHERITANCE_ATTR',
    'DeprecatedNanoContract',
    'FauxImmutable',
    'FauxImmutableMeta',
    'OnChainBlueprint',
    'SKIP_VALIDATION_ATTR',
    '__set_faux_immutable__',
    'create_with_shell',
]



#  Copyright 2023 Hathor Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
from pathlib import Path
from typing import Any, Optional, TypeVar, Union

import yaml
from pydantic import BaseModel

from hathorlib.utils.dict import deep_merge

_EXTENDS_KEY = 'extends'

T = TypeVar('T', bound=BaseModel)


def dict_from_yaml(*, filepath: Union[Path, str]) -> dict[str, Any]:
    """Takes a filepath to a yaml file and returns a dictionary with its contents."""
    if not os.path.isfile(filepath):
        raise ValueError(f"'{filepath}' is not a file")

    with open(filepath, 'r') as file:
        contents = yaml.safe_load(file)

        if contents is None:
            return {}

        if not isinstance(contents, dict):
            raise ValueError(f"'{filepath}' cannot be parsed as a dictionary")

        return contents


def dict_from_extended_yaml(*, filepath: Union[Path, str], custom_root: Optional[Path] = None) -> dict[str, Any]:
    """
    Takes a filepath to a yaml file and returns a dictionary with its contents.

    Supports extending another yaml file via the 'extends' key in the file. The 'extends' value can be an absolute path
    to a yaml file, or a path relative to the base yaml file. The custom_root arg can be provided to set a custom root
    for relative paths, taking lower precedence.

    Note: the 'extends' key is reserved and will not be present in the returned dictionary.
    To opt-out of the extension feature, use dict_from_yaml().
    """
    extension_dict = dict_from_yaml(filepath=filepath)
    file_to_extend = extension_dict.pop(_EXTENDS_KEY, None)

    if not file_to_extend:
        return extension_dict

    filepath_to_extend = Path(filepath).parent / str(file_to_extend)

    if not os.path.isfile(filepath_to_extend) and custom_root:
        filepath_to_extend = custom_root / str(file_to_extend)

    try:
        dict_to_extend = dict_from_extended_yaml(filepath=filepath_to_extend, custom_root=custom_root)
    except RecursionError as e:
        raise ValueError('Cannot parse yaml with recursive extensions.') from e

    extended_dict = deep_merge(dict_to_extend, extension_dict)

    return extended_dict


def model_from_extended_yaml(model: type[T], *, filepath: str, custom_root: Optional[Path] = None) -> T:
    """Takes a pydantic model and a filepath to a yaml file and returns a validated model instance."""
    settings_dict = dict_from_extended_yaml(filepath=filepath, custom_root=custom_root)

    return model.model_validate(settings_dict)



#  Copyright 2025 Hathor Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import Callable, TypeVar

from typing_extensions import ParamSpec

# special attrs:
SKIP_VALIDATION_ATTR: str = '__skip_faux_immutability_validation__'
ALLOW_INHERITANCE_ATTR: str = '__allow_faux_inheritance__'
ALLOW_DUNDER_ATTR: str = '__allow_faux_dunder__'


def _validate_faux_immutable_meta(name: str, bases: tuple[type, ...], attrs: dict[str, object]) -> None:
    """Run validations during faux-immutable class creation."""
    required_attrs = frozenset({
        '__slots__',
    })

    for attr in required_attrs:
        if attr not in attrs:
            raise TypeError(f'faux-immutable class `{name}` must define `{attr}`')

    custom_allowed_dunder_value: tuple[str, ...] = attrs.pop(ALLOW_DUNDER_ATTR, ())  # type: ignore[assignment]
    custom_allowed_dunder = frozenset(custom_allowed_dunder_value)
    allowed_dunder = frozenset({
        '__module__',
        '__qualname__',
        '__doc__',
        '__init__',
        '__call__',
        '__firstlineno__',
        '__static_attributes__',
    }) | custom_allowed_dunder

    # pop the attribute so the created class doesn't have it and it isn't inherited
    allow_inheritance = attrs.pop(ALLOW_INHERITANCE_ATTR, False)

    # Prohibit all other dunder attributes/methods.
    for attr in attrs:
        if '__' in attr and attr not in required_attrs | allowed_dunder:
            raise TypeError(f'faux-immutable class `{name}` must not define `{attr}`')

    # Prohibit inheritance on faux-immutable classes, this may be less strict in the future,
    # but we may only allow bases where `type(base) is FauxImmutableMeta`.
    if len(bases) != 1:
        raise TypeError('faux-immutable only allows one base')

    base, = bases
    if base is not FauxImmutable and not allow_inheritance:
        raise TypeError(f'faux-immutable class `{name}` must inherit from `FauxImmutable` only')


class FauxImmutableMeta(type):
    """
    A metaclass for faux-immutable classes.
    This means the class objects themselves are immutable, that is, `__setattr__` always raises AttributeError.
    Don't use this metaclass directly, inherit from `FauxImmutable` instead.
    """
    __slots__ = ()

    def __new__(cls, name, bases, attrs, **kwargs):
        # validations are just a sanity check to make sure we only apply this metaclass to classes
        # that will actually become immutable, for example, using this metaclass doesn't provide
        # complete faux-immutability if the class doesn't define `__slots__`.
        if not attrs.get(SKIP_VALIDATION_ATTR, False):
            _validate_faux_immutable_meta(name, bases, attrs)
        return super().__new__(cls, name, bases, attrs, **kwargs)

    def __setattr__(cls, name: str, value: object) -> None:
        raise AttributeError(f'cannot set attribute `{name}` on faux-immutable class')


class FauxImmutable(metaclass=FauxImmutableMeta):
    """
    Utility superclass for creating faux-immutable classes.
    Simply inherit from it to define a faux-immutable class.
    """
    __slots__ = ()
    __skip_faux_immutability_validation__: bool = True  # Skip validation to bypass the no dunder rule.

    def __setattr__(self, name: str, value: object) -> None:
        raise AttributeError(f'cannot set attribute `{name}` on faux-immutable object')


T = TypeVar('T', bound=FauxImmutable)
P = ParamSpec('P')


def create_with_shell(cls: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
    """Mimic `cls.__call__` method behavior, but wrapping the created instance with an ad-hoc shell class."""
    # Keep the same name as the original class.
    assert isinstance(cls, type)
    name = cls.__name__

    # The original class is the shell's only base.
    bases = (cls,)

    # The shell doesn't have any slots and must skip validation to bypass the inheritance rule.
    attrs = dict(__slots__=(), __skip_faux_immutability_validation__=True)

    # Create a dynamic class that is only used on this call.
    shell_type: type[T] = type(name, bases, attrs)

    # Use it to instantiate the object, init it, and return it. This mimics the default `__call__` behavior.
    obj: T = cls.__new__(shell_type)  # type: ignore[call-overload]
    shell_type.__init__(obj, *args, **kwargs)
    return obj


def __set_faux_immutable__(obj: FauxImmutable, name: str, value: object) -> None:
    """
    When setting attributes on the `__init__` method of a faux-immutable class,
    use this utility function to bypass the protections.
    Only use it when you know what you're doing.
    """
    if name.startswith('__') and not name.endswith('__'):
        # Account for Python's name mangling.
        name = f'_{obj.__class__.__name__}{name}'

    # This shows that a faux-immutable class is never actually immutable.
    # It's always possible to mutate it via `object.__setattr__`.
    object.__setattr__(obj, name, value)



# Copyright 2023 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from hathorlib.headers.base import VertexBaseHeader
from hathorlib.headers.types import VertexHeaderId
from hathorlib.utils import int_to_bytes, unpack

if TYPE_CHECKING:
    from hathorlib.base_transaction import BaseTransaction
    from hathorlib.transaction import Transaction


@dataclass(frozen=True)
class FeeHeaderEntry:
    token_index: int
    amount: int


@dataclass(frozen=True)
class FeeEntry:
    token_uid: bytes
    amount: int


@dataclass(frozen=True)
class FeeHeader(VertexBaseHeader):
    tx: Transaction
    fees: list[FeeHeaderEntry]

    @classmethod
    def deserialize(cls, tx: BaseTransaction, buf: bytes) -> tuple[FeeHeader, bytes]:
        from hathorlib.base_transaction import bytes_to_output_value

        header_id, buf = buf[:1], buf[1:]
        assert header_id == VertexHeaderId.FEE_HEADER.value

        fees: list[FeeHeaderEntry] = []
        (fees_len,), buf = unpack('!B', buf)

        for _ in range(fees_len):
            (token_index,), buf = unpack('!B', buf)
            amount, buf = bytes_to_output_value(buf)
            fees.append(FeeHeaderEntry(
                token_index=token_index,
                amount=amount,
            ))
        from hathorlib.transaction import Transaction
        assert isinstance(tx, Transaction)
        return cls(
            tx=tx,
            fees=fees,
        ), bytes(buf)

    def serialize(self) -> bytes:
        from hathorlib.base_transaction import output_value_to_bytes

        ret = [
            VertexHeaderId.FEE_HEADER.value,
            int_to_bytes(len(self.fees), 1)
        ]

        for fee in self.fees:
            ret.append(int_to_bytes(fee.token_index, 1))
            ret.append(output_value_to_bytes(fee.amount))

        return b''.join(ret)

    def get_sighash_bytes(self) -> bytes:
        return self.serialize()

    def get_fees(self) -> list[FeeEntry]:
        return [
            FeeEntry(
                token_uid=self.tx.get_token_uid(fee.token_index),
                amount=fee.amount
            )
            for fee in self.fees
        ]



from pathlib import Path

from hathorlib.conf.get_settings import HathorSettings

parent_dir = Path(__file__).parent

MAINNET_SETTINGS_FILEPATH = str(parent_dir / 'mainnet.yml')
TESTNET_INDIA_SETTINGS_FILEPATH = str(parent_dir / 'testnet.yml')
NANO_TESTNET_SETTINGS_FILEPATH = str(parent_dir / 'nano_testnet.yml')
LOCALNET_SETTINGS_FILEPATH = str(parent_dir / 'localnet.yml')
UNITTESTS_SETTINGS_FILEPATH = str(parent_dir / 'unittests.yml')

__all__ = [
    'MAINNET_SETTINGS_FILEPATH',
    'TESTNET_INDIA_SETTINGS_FILEPATH',
    'NANO_TESTNET_SETTINGS_FILEPATH',
    'LOCALNET_SETTINGS_FILEPATH',
    'UNITTESTS_SETTINGS_FILEPATH',
    'HathorSettings',
]
